package core

import (
	"net/url"
	"path"
	"strings"

	"github.com/grufwub/go-errors"
)

// getRequestPaths points to either of the getRequestPath____ functions
var getRequestPath func(string) *Path

// ParseURLEncodedRequest takes a received string and safely parses a request from this
func ParseURLEncodedRequest(received string) (*Request, *errors.Error) {
	// Check for ASCII control bytes
	for i := 0; i < len(received); i++ {
		if received[i] < ' ' || received[i] == 0x7f {
			return nil, errors.NewError(InvalidRequestErr)
		}
	}

	// Split into 2 substrings by '?'. URL path and query
	rawPath, query := SplitBy(received, "?")

	// Unescape path
	rawPath, err := url.PathUnescape(rawPath)
	if err != nil {
		return nil, errors.WrapError(InvalidRequestErr, err)
	}

	// Return new request
	return &Request{getRequestPath(rawPath), query}, nil
}

// ParseInternalRequest parses an internal request string based on the current directory
func ParseInternalRequest(p *Path, line string) *Request {
	rawPath, query := SplitBy(line, "?")
	if path.IsAbs(rawPath) {
		return &Request{getRequestPath(rawPath), query}
	}
	return &Request{newSanitizedPath(p.Root(), rawPath), query}
}

// getRequestPathUserSpacesEnabled creates a Path object from raw path, converting ~USER to user subdirectory roots, else at server root
func getRequestPathUserSpacesEnabled(rawPath string) *Path {
	if userPath := strings.TrimPrefix(rawPath, "/"); strings.HasPrefix(userPath, "~") {
		// We found a user path! Split into the user part, and remaining path
		user, remaining := SplitBy(userPath, "/")

		// Empty user, we been duped! Return server root
		if len(user) <= 1 {
			return &Path{Root, "", "/"}
		}

		// Get sanitized user root, else return server root
		root, ok := sanitizeUserRoot(path.Join("/home", user[1:], userSpace))
		if !ok {
			return &Path{Root, "", "/"}
		}

		// Build new Path
		rel := sanitizeRawPath(root, remaining)
		sel := "/~" + user[1:] + formatSelector(rel)
		return &Path{root, rel, sel}
	}

	// Return regular server root + rawPath
	return newSanitizedPath(Root, rawPath)
}

// getRequestPathUserSpacesDisabled creates a Path object from raw path, always at server root
func getRequestPathUserSpacesDisabled(rawPath string) *Path {
	return newSanitizedPath(Root, rawPath)
}
