package gemini

import (
	"gophi/core"
	"os"
	"strings"

	"github.com/grufwub/go-errors"
)

// serve is the global gemini server's serve function
func serve(client *core.Client) {
	// Receive line from client
	received, err := client.Conn().ReadLine()
	if err != nil {
		client.LogError(clientReadFailStr)
		handleError(client, err)
		return
	}

	// Strip everything up to first '/'
	proto, line := core.SplitBy(string(received), "://")

	// If line is empty, means none supplied --> assume gemini.
	// Else, we check for valid proto
	if line == "" {
		line = proto
	} else if proto != "gemini" {
		client.LogError(clientInvalidRequestStr)
		handleError(client, errors.NewError(invalidProtocolErr))
		return
	}

	// If doesn't start with '/', is providing a host(:port) and we need
	// to make sure this is equal to our current server
	if !strings.HasPrefix(line, "/") {
		var hostPort string
		hostPort, line = core.SplitBy(line, "/")
		host, port := core.SplitBy(hostPort, ":")
		if host != core.Hostname || (port != "" && port != core.Port) {
			client.LogError(clientInvalidRequestStr)
			handleError(client, errors.NewError(invalidHostPortErr))
			return
		}
	}

	// Parse new request
	request, err := core.ParseURLEncodedRequest(line)
	if err != nil {
		client.LogError(clientRequestParseFailStr)
		handleError(client, err)
		return
	}

	// Handle the request!
	err = core.HandleClient(client, request)

	// Final error handling
	if err != nil {
		handleError(client, err)
		client.LogError(clientServeFailStr, request.String())
	} else {
		client.LogInfo(clientServedStr, request.String())
	}
}

// handleError determines whether to send an error response to the client, and logs to system
func handleError(client *core.Client, err *errors.Error) {
	response, ok := generateErrorResponse(err)
	if ok {
		client.Conn().Write(response)
	}
	core.SystemLog.Errorf(err.Error())
}

func handleDirectory(client *core.Client, fd *os.File, p *core.Path) *errors.Error {
	// First check for index gem, create gem Path object
	indexGem := p.JoinPath("index.gmi")

	// If index gem exists, we fetch this
	fd2, err := core.OpenFile(indexGem)
	if err == nil {
		stat, osErr := fd2.Stat()
		if osErr == nil {
			// Fetch gem and defer close
			defer fd2.Close()
			return core.FetchFile(client, fd2, stat, indexGem)
		}

		// Else, just close fd2
		fd2.Close()
	}

	// Directory page
	dirContents := ""

	// Add directory heading + empty line
	dirContents += "# [ " + core.Hostname + p.Selector() + " ]\n"
	dirContents += "\n"

	// Scan directory and build lines
	err = core.ScanDirectory(
		fd,
		p,
		func(file os.FileInfo, fp *core.Path) {
			// Append new formatted file listing
			dirContents += "=> " + fp.Selector() + " * " + fp.Selector() + "\n"
		},
	)
	if err != nil {
		return err
	}

	// Generate gem file header
	header := buildResponseHeader("20", gemMimeType)

	// Write contents!
	return client.Conn().Write(append(header, []byte(dirContents)...))
}

func handleLargeFile(client *core.Client, file *os.File, p *core.Path) *errors.Error {
	// Build the response header
	header := buildResponseHeader("20", getFileStatusMeta(p))

	// Write the initial header (or return!)
	err := client.Conn().Write(header)
	if err != nil {
		return err
	}

	// Finally write directly from file
	return client.Conn().ReadFrom(file)
}

// newFileContents returns a new FileContents object
func newFileContent(p *core.Path) core.FileContent {
	return &headerPlusFileContent{}
}
